/**
 * \file CplxFFT.h
 * \brief performs complex FFT
 * \author Maciek Smuga-Otto <maciek@ssec.wisc.edu>
 *
 * based directly on the Scanning-HIS rsh2fbf application, written by Ray Garcia <rayg@ssec.wisc.edu>
 *
 *  \version $Id: CplxFFT.hxx,v 1.8 2005/11/21 00:07:28 rayg Exp $
 *
 *  \par Copyright:
 *  \verbatim
 *
 *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
 *  Space Science and Engineering Center
 *  University of Wisconsin - Madison, USA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *  \endverbatim
 */
 
 
/* Interface specification
Input port: obs_in
    obs_in->cplxinterferogram
    
Output port: obs_out
    obs_out->cplxspectrum
*/

#ifndef H_CPLXFFT
#define H_CPLXFFT

#ifndef NO_DEPENDENCY_INCLUDES

// the FFTW system-dependent #include polka
#ifdef __APPLE__
#include "dfftw_threads.h"
#include "drfftw_threads.h"
#elif defined(__CYGWIN32__)
#include "fftw.h"
#include "rfftw.h"
#define fftw_threads_init() 
#define fftw_threads_one(a,b,c,d) fftw_one(b,c,d)
#define rfftw_threads_one(a,b,c,d) rfftw_one(b,c,d)
#else
#include "fftw_threads.h"
#include "rfftw_threads.h"
#endif

// GIPS includes
#include "Physics.h"
#include "Instrument.h"
#include "ProtocolRole.h"
#include "Operator.h"

#include "Types.h"

#endif


/* Interface specification; refer to data dictionary for specifics 

Interferograms and spectra are accessed using std::vector<> semantics

CplxFFT<Traits,Ports>:
    Input port: obs_in
        obs_in->cplxinterferogram
    Output port: obs_out
        obs_out->cplxspectrum
*/


namespace gips {

template <typename Traits, typename Ports>
class CplxFFT 
{
protected:
    typedef BufferTypes<Traits> Types;

    enum {  shiftedInterferogramSize = Types::InterferogramSize,
            doubledRawSpectrumSize = Types::FftBufferSize };

    // primary input and output connections to the external system
    Ports P;

    fftw_plan plan; // cached machine-dependent plan - should it be const?

    // for non-complex-doubles, we need an intermediate copy, ugh
    template<typename FromType>
    void anyFFT( std::complex<double> *dest, const FromType *src )
    {
        std::complex<double> tmp[ Types::FftBufferSize ];
        std::copy( src, src+Types::FftBufferSize, &tmp[0] );
        fftw_one( plan, (fftw_complex *)&tmp[0], (fftw_complex *)dest );
        // FIXME: issue a monitoring warning that extraneous copies are being used
    }
    
    // specialize to the optimal case
    void anyFFT( std::complex<double> *dest, std::complex<double> *src )
    {
        fftw_one( plan, (fftw_complex *)src, (fftw_complex *)dest );
    }

public:
    explicit CplxFFT(const Ports &p): P(p)
    {
        plan = fftw_create_plan( Types::FftBufferSize, FFTW_FORWARD, FFTW_MEASURE ); 
        // alternatively, use FFTW_ESTIMATE instead of FFTW_MEASURE for faster startup

        fftw_threads_init();    
    }; // ctor

    ~CplxFFT() 
    {
        fftw_destroy_plan( plan );
    }; // dtor

    Operator::Result operator() ( )
    {
        anyFFT( &P.obs_out->cplxspectrum[0], &P.obs_in->cplxinterferogram[0] );
        return Operator::OK;
    };

}; // class CplxFFT



template< typename Traits >
struct CplxFFTInterface
{
    typedef BufferTypes<Traits> Types;

    struct obs_in_t 
    {
        typename Types::Interferogram cplxinterferogram;
    };
    
    struct obs_out_t 
    {
        typename Types::FftSpectrum cplxspectrum;
    };
    
    struct Ports 
    {
        // simplest connectors -- pointers.
        typedef const obs_in_t *obs_in_p;
        typedef obs_out_t *obs_out_p;
        
        obs_in_p obs_in;
        obs_out_p obs_out;
        
        Ports(obs_in_p obs_in_, obs_out_p obs_out_): 
            obs_in(obs_in_), obs_out(obs_out_) { }      
    };

};

    

    
} // namespace
#endif
    
    

