/**
 * \file IfgShift.h
 * \brief Shifts raw interferogram to form expected by the FFT operator.
 * \author Maciek Smuga-Otto <maciek@ssec.wisc.edu>
 *
 * based directly on the Scanning-HIS rsh2fbf application, written by Ray Garcia <rayg@ssec.wisc.edu>
 *
 *  \version $Id: IfgShift.hxx,v 1.6 2005/10/27 21:51:32 rayg Exp $
 *
 *  \par Copyright:
 *  \verbatim
 *
 *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
 *  Space Science and Engineering Center
 *  University of Wisconsin - Madison, USA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *  \endverbatim
 */

/* Interface specification; refer to data dictionary for specifics 

Interferograms are accessed using std::vector<> semantics

IfgShift<Traits,Ports>:
    Input port: obs_in
        obs_in->interferogram
    Output port: obs_out
        obs_out->interferogram

CplxIfgShift<Traits,Ports>:
    Input port: obs_in
        obs_in->cplxinterferogram        
    Output port: obs_out
        obs_out->cplxinterferogram

RealToCplxIfgShift<Traits,Ports>:
    Input port: obs_in
        obs_in->interferogram        
    Output port: obs_out
        obs_out->cplxinterferogram
 
*/
 
// 


#ifndef H_IFGSHIFT
#define H_IFGSHIFT

#ifndef NO_DEPENDENCY_INCLUDES
// GIPS includes
#include "Physics.h"
#include "Instrument.h"
#include "ProtocolRole.h"
#include "Operator.h"
#include "Types.h"
#include <complex>

#endif

// NOTE: this file will be automatically generated by a flanging compiler in the future.
// FIXME: parameterize on the protocols!
namespace gips {


template< typename A, typename B >
void anyShift( B &out, const A &in )
{
    unsigned N(in.size()), halfN(N/2); 
    unsigned src(0), dest(0);
    
    for( src = halfN; src<N; src++,dest++ ) 
        out[dest] = in[src];

    for( src=0; src<halfN; src++, dest++ ) 
        out[dest] = in[src]; 
}


template< typename Traits, typename Ports >
class IfgShift
{    
protected:
    // use the contents of the traits structure to configure the class
    typedef BufferTypes<Traits> Types;
            
    // instance variable containing a copy of the connectors
    Ports P;    
    
public:
    // Initialize the connectors from the provided spec. 
    // One advantage of this is that it semantically hands off the connection configuration
    // such that reconfiguration is not allowed. 
    // Note that the 'settings' structure has been discarded in favor of info derived from
    // the pipeline Traits. 
    
    explicit IfgShift(const Ports &p): P(p)
    {
    }
    
    ~IfgShift() 
    { 
    }
    
    // The meat of the template. 
    Operator::Result operator() ( )
    {
        assert( P.obs_in->interferogram.size() == P.obs_out->interferogram.size() );
        //std::cerr << "C++: in IfgShiftOperator::operator()" << std::endl;
    
        // Block of interface code, getting access to ports.
        //const settings_t& settings(ports.settings.getForInput());
        anyShift( P.obs_out->interferogram, P.obs_in->interferogram );
        
        return Operator::OK;    
    } 
}; // class 


template< typename Traits, typename Ports >
class CplxIfgShift
{    
protected:
    // use the contents of the traits structure to configure the class
    typedef BufferTypes<Traits> Types;
            
    // instance variable containing a copy of the connectors
    Ports P;    
    
public:
    // Initialize the connectors from the provided spec. 
    // One advantage of this is that it semantically hands off the connection configuration
    // such that reconfiguration is not allowed. 
    // Note that the 'settings' structure has been discarded in favor of info derived from
    // the pipeline Traits. 
    
    explicit CplxIfgShift(const Ports &p): P(p)
    {
    }
    
    ~CplxIfgShift() 
    { 
    }
    
    // The meat of the template. 
    Operator::Result operator() ( )
    {
        assert( P.obs_in->cplxinterferogram.size() == P.obs_out->cplxinterferogram.size() );

        //std::cerr << "C++: in IfgShiftOperator::operator()" << std::endl;
    
        // Block of interface code, getting access to ports.
        //const settings_t& settings(ports.settings.getForInput());
        anyShift( P.obs_out->cplxinterferogram, P.obs_in->cplxinterferogram );
        
        return Operator::OK;    
    } 
}; // class 


template< typename Traits, typename Ports >
class RealToCplxIfgShift
{    
protected:
    // use the contents of the traits structure to configure the class
    typedef BufferTypes<Traits> Types;
            
    // instance variable containing a copy of the connectors
    Ports P;    
    
public:
    // Initialize the connectors from the provided spec. 
    // One advantage of this is that it semantically hands off the connection configuration
    // such that reconfiguration is not allowed. 
    // Note that the 'settings' structure has been discarded in favor of info derived from
    // the pipeline Traits. 
    
    explicit RealToCplxIfgShift(const Ports &p): P(p)
    {
    }
    
    ~RealToCplxIfgShift() 
    { 
    }
    
    // The meat of the template. 
    Operator::Result operator() ( )
    {
        assert( P.obs_in->interferogram.size() == P.obs_out->cplxinterferogram.size() );
        //std::cerr << "C++: in IfgShiftOperator::operator()" << std::endl;
    
        // Block of interface code, getting access to ports.
        //const settings_t& settings(ports.settings.getForInput());
        anyShift( P.obs_out->cplxinterferogram, P.obs_in->interferogram );
        
        return Operator::OK;    
    } 
}; // class 



///////////////////////////////////////////////////////////////////
// sample protocols for testing.
// In this case, we're using regular pointers as connectors instead of anything more exotic.
// These should ultimately be deprecated in favor of a 
// formal interface specification usable by a flanging compiler,
// so that testing protocols can be generated.
// Interface is only dependent on Traits and provides a Ports
// e.g. IfgShift< mytraits, IfgShiftInterface<mytraits>::Ports > for testing
// this could stand some improvement.

template< typename Traits >
struct IfgShiftInterface 
{
    typedef BufferTypes<Traits> Types;
    struct obs_in_t
    {
        // interferogram conforms to STL read-only vector access
        typename Types::Interferogram cplxinterferogram;
    };
    
    struct obs_out_t
    {
        typename Types::Interferogram cplxinterferogram;
    };
    
    
    struct Ports 
    {
        typedef const obs_in_t* obs_in_p;
        typedef obs_out_t* obs_out_p;
        // elementary connectors - pointers. 
        obs_in_p  obs_in;
        obs_out_p obs_out;
    
        Ports(obs_in_p obs_in_, obs_out_p obs_out_): 
            obs_in(obs_in_), obs_out(obs_out_) { }
    };
};




} // namespace





#endif
    
