
/** 
*  \file PiecewiseLinearCalibrationContext.hxx
*  \brief Template for modeling blackbody temperatures as linear windows in time.
*
*  \author R.K.Garcia <rayg@ssec.wisc.edu>
*
*  \version $Id: PiecewiseLinearCalibrationContext.hxx,v 1.4 2005/11/21 02:33:16 rayg Exp $
*
*  \par Copyright:
*  \verbatim
*
*  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
*  Space Science and Engineering Center
*  University of Wisconsin - Madison, USA
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*  \endverbatim
*/

/* Interface specification

PiecewiseLinearCalibrationStudy<Traits,Ports>
    
    obs_in->cplxspectrum[] : std::vector<>-like blackbody or other spectrum 
    obs_in->time : double-like epoch time for observation
    obs_in->scenetype : char scene type: 'H', 'A', 'Z', 'E'
    obs_in->scandir : char scan direction: 'F', 'B'
    
    cache_out->context_cache{} : std::map<>-like context cache holding opaque type
    
ThreeBodyPolynomialCalibrationSynthesis<Traits,Ports>

    cache_in->context_cache{} : std::map<>-like context cache of PolynomialCalibrationContext<Traits> objects
    
    obs_in->time : desired observation time
    obs_in->scandir : scan direction

    zbb_out->cplxspectrum[] 
    abb_out->cplxspectrum[]
    hbb_out->cplxspectrum[]
        
*/

#ifdef DEBUG
#define DBG(x) x
#else
#define DBG(x) 
#endif

#include <algorithm>
#include <iostream>
#include <map> 

#include "Types.h"

namespace gips 
{




/****** 
template< typename Traits >
class AbstractCalibrationContext
{
protected:
    typedef BufferTypes<Traits> Types;
public:
    typedef Types::Spectrum Spectrum;    
    virtual void operator()( Spectrum &X, double T ) = 0;
    virtual ~AbstractCalibrationContext() { }

// information on how to cache these things quickly and painlessly
    typedef boost::shared_ptr< AbstractCalibrationContext<Traits> > context_p;

    typedef std::pair< char, char > k_scene; // H/A/Z, F/B
    typedef std::pair< k_scene, double > k_cache; // scene, start time 
    typedef std::map< k_cache, context_p > BasicCacheType;
};
******/



/*********    
template< typename Traits >
class PLC_Context: public AbstractCalibrationContext<Traits>
{
protected:
    double beginT;
    Spectrum slope, intercept;
public:
    PLC_Context( double beginT_, double endT_,
                const Spectrum &begin, 
                const Spectrum &end ): beginT(beginT_)
    {
        // t0 == 0.0, we always subtract off beginT when using this contextual frame
        double t1 = endT_ - beginT_;
        for( unsigned i=0; i < begin.size(); ++i )
        {
            slope[i] = (end[i] - begin[i]) / t1;
            intercept[i] = begin[i];
        }
    }    

    virtual void operator()( Spectrum &X, double T )
    {
        double t( T - beginT );
        for( unsigned i=0; i < X.size(); ++i )
            X[i] = slope[i] * t + intercept[i];
    }
};
*******/


// FIXME: Promote this to outside PLCC.hxx!
template< typename Traits, unsigned Order = 3 >
class PolynomialCalibrationContext
{
protected:
    enum { COEFFS = Order + 1 };
    typedef BufferTypes<Traits> Types;
    
public:
    typedef typename Types::Spectrum Spectrum;
    double baseT;
    Spectrum A[COEFFS];   
    
    void clear() 
    {
        baseT = 0.0;
        for( unsigned c=0; c<COEFFS; c++ )
            std::fill( A[c].begin(), A[c].end(), 0 );
    }
    
    PolynomialCalibrationContext()
    {
        clear();
    }
        
    template< typename Target >
    void operator()( Target &X, double T )
    {
        assert( X.size() == A[0].size() );
        
        T -= baseT;        
        for( unsigned w=0; w < X.size(); ++w )
        {
            typename Target::value_type Y = T * A[1][w] + A[0][w];
            for( unsigned c=2; c < COEFFS; ++c )
            {
                double a( A[c][w] );
                if (a!=0.0)
                    Y += pow(T,c) * a;
            }
            X[w] = Y;
        }
    }

// information on how to cache these things quickly and painlessly
// unless you have a more involved way to deal with storing/serializing/deserializing

    typedef std::pair< char, char > k_scene; // H/A/Z, F/B
    typedef std::pair< k_scene, double > k_cache; // scene, start time 

    static k_cache key( char scenetype, char scandir, double time ) 
    {
        return k_cache( k_scene( scenetype, scandir ), time );
    }

    typedef std::map< k_cache, PolynomialCalibrationContext > BasicCacheType;
    
    template<typename Iteratorlike>    
    inline static bool sanityCheckCacheEntry( char scenetype, char scandir, Iteratorlike x )
    {
        DBG( std::clog << "sanity check: " << x->first.first.first << x->first.first.second << 'v' << scenetype << scandir << std::endl );
        // make sure that we're returning something that actually is ffor this scene and direction!
        return bool(x->first.first == k_scene(scenetype,scandir));
    }

};




template< typename Traits, typename Ports >
class PiecewiseLinearCalibrationStudy
{
protected:
    typedef BufferTypes<Traits> Types;
    typedef PolynomialCalibrationContext<Traits> PCC;

    Ports P;
    
    // all state for this capsule is in the cache!
    
    // start is the previous calibration frame, which has previous time, and previous spectrum in A[0]
    void calculate_entry( PCC &out, const PCC &start ) 
    {
        out.clear();
        
        // starting time is 0.0 - (obs_in->time - prevT)
        // ending time is 0.0, corresponding to obs_in->time
        // This makes it trivial to recover the spectrum at obs_in->time from the PCC
        
        double deltaT = P.obs_in->time - start.baseT;
        out.baseT = P.obs_in->time;
        
        for( unsigned w=0; w < P.obs_in->cplxspectrum.size(); ++w )
        {
            out.A[0][w] = P.obs_in->cplxspectrum[w];
            out.A[1][w] = (P.obs_in->cplxspectrum[w] - start.A[0][w]) / deltaT;
        }
    }

    // create initial cache entry going back to T=0, zeroed A[1..n]
    void make_initial_entry( PCC &out )
    {
        out.clear();
        out.baseT = P.obs_in->time;
        //out.A[0] = P.obs_in->cplxspectrum;
        for( unsigned w=0; w < P.obs_in->cplxspectrum.size(); ++w )
            out.A[0][w] = P.obs_in->cplxspectrum[w];
        DBG( std::clog << "initializing for " << P.obs_in->scenetype << P.obs_in->scandir << std::endl );
    }
    
    template<typename CacheIterator, typename CacheKey >
    int update( CacheIterator iter, const CacheKey &kiki )
    {
        // if the iterator isn't the start of the whole cache,
        // and it's not a perfect match for the data, 
        // go to the previous entry
        // since lower_bound returns the iterator pointing to the first element with a key not less than k,
        // but what we want is the last key which is less than or equal to k!
        if ( (iter != P.cache_out->context_cache.begin()) && (iter->first != kiki) )
            --iter;
    
        if ( (iter == P.cache_out->context_cache.end()) || 
                !PCC::sanityCheckCacheEntry(P.obs_in->scenetype,P.obs_in->scandir,iter) )
        {
            make_initial_entry( P.cache_out->context_cache[ PCC::key(P.obs_in->scenetype, P.obs_in->scandir, 0) ] );
        }
        else
        {
            // use the entry at *iter as the starting point for a new context
            // base time of that context is the starting time (and thus used in the key, so lower_bound works)
            calculate_entry( P.cache_out->context_cache[ PCC::key(P.obs_in->scenetype, P.obs_in->scandir, iter->second.baseT) ], iter->second );
        }
        return 0;
    }
    
    template< typename CacheKey > // is a little bit silly... okay more than a little bit
    int update( const CacheKey &kiki )
    {
        return update( P.cache_out->context_cache.lower_bound( kiki ), kiki );
    }

public:

    PiecewiseLinearCalibrationStudy( const Ports &p_ ): P(p_) { }

    int operator()( ) 
    {
        // recover most recent entry for this scene type from cache
        // create a new one based on it; if it doesn't exist, initialize for this scene type
        return update( PCC::key(P.obs_in->scenetype, P.obs_in->scandir, P.obs_in->time) );
        
        return 0;  // FIXME: return failure if we get a division by zero 
    }    
};


template< typename Traits, typename Ports >
class ThreeBodyPolynomialCalibrationSynthesis
{
protected:
    typedef BufferTypes<Traits> Types;
    typedef PolynomialCalibrationContext<Traits> PCC;

    Ports P;
    //const PCC nullContext;

    class noContextAvailable
    {
    };

    template< typename Iteratorlike >
    const PCC &sanitized( char scenetype, char scandir, Iteratorlike x )
    {
        // make sure that we're returning something that actually is for this scene and direction!
        if (!PCC::sanityCheckCacheEntry( scenetype,scandir,x )) 
        {
            throw noContextAvailable();
            /// FIXME: WARNING, exception constrained within the capsule or other assertion
            //return nullContext;
        }
        return *x;
    }

    inline const PCC &invokeBestContextFor( char scenetype )
    {
        // find the best context to use by searching the database
        return sanitized( scenetype, obs_in->scandir, 
                cache_in->context_cache.lower_bound( PCC::key(scenetype, obs_in->scandir, obs_in->time) ) );
    }
    
    // FIXME:
    // Do we interpolate blackbody temperatures as well?
    // Do we use the interpolated blackbody temperature as compared to the actual as an error metric?
    
    
public:
    ThreeBodyPolynomialCalibrationSynthesis( const Ports &p_ ): P(p_) { }
    
    int operator()( )
    {
        char ctx='-';
        try 
        {
            ctx='H'; invokeBestContextFor(ctx)( hbb_out->cplxspectrum, obs_in->time );
            ctx='A'; invokeBestContextFor(ctx)( abb_out->cplxspectrum, obs_in->time );
            ctx='Z'; invokeBestContextFor(ctx)( zbb_out->cplxspectrum, obs_in->time );
        }
        catch( noContextAvailable nca )
        {
            std::cerr << __FILE__ << ':' << __LINE__ << ": no context found for " << ctx << std::endl;
            return -1;
        }
        return 0;
    }    
};

} // namespace
