/** 
*  \file ProtocolRole.h
*  \brief Protocol Roles for Connectors
*
*  \author M.Smuga-Otto <maciek@ssec.wisc.edu>, R.K.Garcia <rayg@ssec.wisc.edu>
*
*  \version $Id: ProtocolRole.h,v 1.12 2005/10/19 21:11:37 rayg Exp $
*
*  \verbatim
*
*  Connectors provide content buffers for data moving between operators and stages.
*  These archetypes allow system builders to describe the connection requirements
*  of their operators without committing to an implementation (ie they are abstract interfaces).
*  
*  System builders are free to use the existing connector template library to attach
*  the operators, or to implement their own conformant connectors in order to bridge
*  data between second- and third-party systems; for instance, connectors which 
*  allow Python files and numarray variables to act as input to a processing pipeline.
*
*  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
*  Space Science and Engineering Center
*  University of Wisconsin - Madison, USA
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
* 
*  \endverbatim
*/

// break out the Protocol Roles (ie. port interfaces) into a separate include file which can be used
// in building capsule translation units without carrying along the entire weight of the Connector template classes

#ifndef H_PROLE
#define H_PROLE

#ifndef NO_DEPENDENCY_INCLUDES
#include <iosfwd>
#endif


namespace Connector
{

    /**
    ** InputPort<X> : template for a class which provides input to an operator
     **/
    template < typename Payload > 
    class InputPort
    {
    public:
        virtual const Payload& getForInput( ) = 0;
        virtual ~InputPort() { }
    };


    /**
    **  OutputPort<X> : template for a class which provides an output buffer for an operator     
     **/
    template < typename Payload > 
    class OutputPort
    {
    public:
        virtual Payload& getForOutput() = 0;
        virtual ~OutputPort() { }
    };


    /** MonitorPort provides an ostream-like interface 
     * to an external system monitoring performance and error information. 
     * refer to Monitor.h for standard tokens that can be passed into it
     */
    class MonitorPort 
    {
    public:
        virtual operator std::ostream &();
        virtual ~MonitorPort() { }
    };
    
}; // namespace Connector

#endif

