/** 
*  \file RadiometricCalibration.hxx
*  \brief Template for radiometric calibration equation.
*    
*
*  \author R.K.Garcia <rayg@ssec.wisc.edu>
*
*  \version $Id: RadiometricCalibration.hxx,v 1.6 2005/11/22 03:17:43 maciek Exp $
*
*  \par Copyright:
*  \verbatim
*
*  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
*  Space Science and Engineering Center
*  University of Wisconsin - Madison, USA
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*  \endverbatim
*/

/* Interface specification

ThreeBodyRadiometricCalibration<Traits,Ports>

    obs_in->cplxspectrum[] : earth observed spectrum
    
    abb_in->cplxspectrum[] : cold blackbody estimated spectrum matched to obs_in data
    abb_in->bbtemp        : effective cold blackbody temperature at observation time
    abb_in->reflectedtemp : instrument reflected temperature for this blackbody
                            this is likely to be dominated by the flip-in calmirror temperature
                            which is controlled
    
    hbb_in->cplxspectrum[] : hot blackbody estimated spectrum matched to obs_in data
    hbb_in->bbtemp        : effective hot blackbody temperature at observation time
    hbb_in->reflectedtemp : instrument reflected temperature for this blackbody
                            this is likely to be dominated by the flip-in calmirror temperature
                            which is controlled
    
    zbb_in->cplxspectrum[] : space background estimated spectrum matched to obs_in data
    zbb_in->bbtemp        : space 
    
    ref_in->telescopeTau : transmission of the telescope system for earth views
    ref_in->calmirrorTau : transmission of the calibration flip-in mirror system
    ref_in->wavenumbers[] : wavenumber array, optimally as Physics::wnums_t tuple
    ref_in->bbemissivity[] : emissivity term for blackbodies, 0.0-1.0, interpolated to wavenumber scale from ref UW_GIFTS_BB_cavity_emissivity.txt 10/30/05
    Ports::wavenumbers_t  : type of ref_in->wavenumbers
    Ports::emissivity_t : type of ref->in->bbemissivity

    cal_out->spectrum[] : real calibrated earth radiance
    
    diag_out->spectrum[] : diagnostic spectrum, the imaginary leftovers from calibration

TwoBodyRadiometricCalibration<Traits,Ports>

    obs_in->cplxspectrum[] : earth observed spectrum

    ref_in->wavenumbers[] : wavenumber array, optimally as Physics::wnums_t tuple
    Ports::wavenumbers_t  : type of ref_in->wavenumbers
    
    abb_in->cplxspectrum[] : cold blackbody estimated spectrum matched to obs_in data
    abb_in->bbtemp        : cold blackbody temperature at observation time
    
    hbb_in->cplxspectrum[] : hot blackbody estimated spectrum matched to obs_in data
    hbb_in->bbtemp        : hot blackbody temperature at observation time
    
    cal_out->spectrum[] : real calibrated earth radiance
    
    diag_out->spectrum[] : diagnostic spectrum, the imaginary leftovers from calibration

*/
 
#ifndef HXX_RADCAL
#define HXX_RADCAL

#include <vector>
#include <cassert>
#include <cmath>

#include "Types.h"
#include "Planck.hxx"

namespace gips
{

template
<
    typename Traits,
    typename Ports
>
class TwoBodyRadiometricCalibration
{
protected:
    Ports P;
    
public:
    explicit TwoBodyRadiometricCalibration( const Ports &p_ ): P(p_) { }

    void operator()( void )
    {
        assert( P.ref_in->wavenumbers.size() == P.obs_in->cplxspectrum.size() );
        assert( P.ref_in->wavenumbers.size() == P.abb_in->cplxspectrum.size() );
        assert( P.ref_in->wavenumbers.size() == P.hbb_in->cplxspectrum.size() );

        PlanckRadiances<typename Ports::wavenumbers_t> Bc( P.abb_in->bbtemp, P.ref_in->wavenumbers );
        PlanckRadiances<typename Ports::wavenumbers_t> Bh( P.hbb_in->bbtemp, P.ref_in->wavenumbers );        
        
        for( unsigned w = 0; w < P.ref_in->wavenumbers.size(); w++ )
        {
            std::complex<double> rad = (P.obs_in->spectrum[w] - P.abb_in->spectrum[w])
                                        / (P.hbb_in->spectrum[w] - P.abb_in->spectrum[w]);
            cal_out->spectrum[w] = rad.real();
            diag_out->spectrum[w] = rad.imag();
        }
    }
    
}; // template<> class TwoBodyRadiometricCalibration


template
<
    typename Traits,
    typename Ports
>
class ThreeBodyRadiometricCalibration
{
protected:
    Ports P;
        
public:

    explicit ThreeBodyRadiometricCalibration( const Ports &p_ ): P(p_) { }

    void operator()( void )
    {
        assert( P.ref_in->wavenumbers.size() == P.obs_in->cplxspectrum.size() );
        assert( P.ref_in->wavenumbers.size() == P.abb_in->cplxspectrum.size() );
        assert( P.ref_in->wavenumbers.size() == P.hbb_in->cplxspectrum.size() );
        assert( P.ref_in->wavenumbers.size() == P.zbb_in->cplxspectrum.size() );        
        
        // NOTE: the use of interpolated BB temp versus actual BB temp 
        // We want to start by using the actual measurement, though this may need to be filtered due to 
        // possible noise in the BB temp measurement.. Also, may want to use interpolated vs real BB temp 
        // as a quality metric or correction term to the calibration model.
        
        // We salt the planck radiances to get effective planck radiances
        // This B'(wavenumber) = bbemissivity[wavenumber] * B(wavenumber,bbtemp) + (1-bbemissivity[wavenumber]) * B(wavenumber,reflectedtemp)
        // bbemissivity[wavenumber] is ground-characterized reference data computed using monte-carlo model
        // emissivity of space is expected to be 1.0
        
        // NOTE: The blackbody temperatures input into this system are synthesized in the same manner as the 
        // complex BB spectra; therefore the actual blackbody temperature is not relevant to the calibration.
        // The actual blackbody temperature and its deviation from the interpolated/extrapolated/synthesized form
        // may be used as an error term, correction or quality metric of the calibration study model. 
        
        EffectivePlanckRadiances<typename Ports::wavenumbers_t, typename Ports::emissivity_t> Bc( P.abb_in->bbtemp, P.abb_in->reflectedtemp, &P.ref_in->bbemissivity, &P.ref_in->wavenumbers );
        EffectivePlanckRadiances<typename Ports::wavenumbers_t, typename Ports::emissivity_t> Bh( P.hbb_in->bbtemp, P.hbb_in->reflectedtemp, &P.ref_in->bbemissivity, &P.ref_in->wavenumbers );        
        PlanckRadiances<typename Ports::wavenumbers_t> Bs( P.zbb_in->bbtemp, &P.ref_in->wavenumbers );        

        double netTau = (P.ref_in->calmirrorTau / P.ref_in->telescopeTau);
        
        for( unsigned w = 0; w < P.ref_in->wavenumbers.size(); w++ )
        {
            /* N = (Tm/Tt) * (Bh-Bc) * Re( (Ce-Cs) / (Ch-Cc) ) + Bs */            
            
            std::complex<double> CCCC = (P.obs_in->cplxspectrum[w] - P.zbb_in->cplxspectrum[w])
                                        / (P.hbb_in->cplxspectrum[w] - P.abb_in->cplxspectrum[w]);
            double TTBH = netTau * (Bh[w] - Bc[w]);
            
            std::complex<double> rad = CCCC * TTBH + Bs[w];
            
            P.cal_out->spectrum[w] = rad.real();
            P.diag_out->spectrum[w] = rad.imag();
        }
    }
    
}; // template<> class ThreeBodyRadiometricCalibration



}; // namespace gips

#endif
