/** 
*  \file FFTResampler.h
*  \brief Spectral Resampling Algorithm.
*
*  \author M. Smuga-Otto <maciek@ssec.wisc.edu>
*
*  \version $Id: Resample.hxx,v 1.2 2005/11/15 22:04:12 maciek Exp $
*
*  \par Copyright:
*  \verbatim
*
*  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
*  Space Science and Engineering Center
*  University of Wisconsin - Madison, USA
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
* 
* \endverbatim
*
* Interface (pretty simple for this stage):
*
* obs_in->spectrum[] : spectrally uncalibrated data
* ref_in->tau[]
* ref_in->wavenumbers[] : target wavenumbers
* obs_out->spectrum[]: spectrally calibrated data
*
*/

#ifndef H_Resample
#define H_Resample

// C++ STL includes
#include <algorithm>
#include <cmath>
#include <functional>
#include <vector>

// FFTW includes
#include <fftw3.h>

// GIPS includes
//#include "Instrument.h"
//#include "Physics.h"
//#include "ProtocolRole.h"
//#include "Types.h"

namespace gips  {


template <typename T>
class Fftw3RealWrapper
{
    int N;
    double *in, *out;
    fftw_plan p;
    
public:
    Fftw3RealWrapper(int N_, fftw_r2r_kind kind = FFTW_R2HC, int flags = FFTW_ESTIMATE):
	N(N_),
	in((double *)fftw_malloc(sizeof(double) * N)),
	out((double *)fftw_malloc(sizeof(double) * N)),
	p(fftw_plan_r2r_1d(N, in, out, kind, flags)) {}

    ~Fftw3RealWrapper()
	{
	    fftw_destroy_plan(p);
	    fftw_free(in);
	    fftw_free(out);
	}

    void operator()(const std::vector<T>& vin, std::vector<T>& vout)
	{
	    assert(vin.size() == N && vout.size() == N);
	    copy(vin.begin(), vin.end(), in);
	    fftw_execute(p);
	    std::copy(out, out+N, vout.begin());
	}
}; // Fftw3Realwrapper

struct ResampleConfig {
    unsigned int upsampledSize;
    unsigned int spectrumWidth;
};
//static const size_t UPSAMPLED_IFG_SIZE = 65536;
//static const size_t UPSAMPLED_SPECTRUM_SIZE = 32769; // this will be == upsampledSize    

class ResampleSpectrum 
{   

    const ResampleConfig config;
    const unsigned int UPSAMPLED_IFG_SIZE;
    const unsigned int UPSAMPLED_SPECTRUM_SIZE;    

    const unsigned int SPECTRUM_SIZE;
    const unsigned int IFG_SIZE;
    const double NORMALIZATION;

    std::vector<double> spec_reflected;
    std::vector<double> short_ifg;
    
    std::vector<double> ifg_zerofilled;
    std::vector<double> spec_upsampled_reflected;
    std::vector<double> spec_upsampled;
    
    Fftw3RealWrapper<double> fft;
    Fftw3RealWrapper<double> fftback;

        
public: 
    explicit ResampleSpectrum(ResampleConfig config_):
	config(config_),
	UPSAMPLED_IFG_SIZE((config.upsampledSize-1) * 2),
	UPSAMPLED_SPECTRUM_SIZE(config.upsampledSize),
	SPECTRUM_SIZE( config.spectrumWidth ),
	IFG_SIZE( (SPECTRUM_SIZE - 1) * 2 ),
	NORMALIZATION( 1.0 /IFG_SIZE ),
	spec_reflected( IFG_SIZE, 0.0 ),
	short_ifg( IFG_SIZE, 0.0 ),
	ifg_zerofilled( UPSAMPLED_IFG_SIZE, 0.0 ),
	spec_upsampled_reflected( UPSAMPLED_IFG_SIZE, 0.0 ),
	spec_upsampled( UPSAMPLED_SPECTRUM_SIZE, 0.0 ),
	fft(IFG_SIZE),
	fftback(UPSAMPLED_IFG_SIZE)
        {}
    
    ~ResampleSpectrum(){}
        
private:        
    /***********************************
    upsample incoming spectrum
    ***********************************/
    void upsample( std::vector<double> &spec_upsampled, 
                                const std::vector<double>& spec_in)
    {
        // reflect the spectrum
        std::copy( spec_in.begin(), spec_in.end(), spec_reflected.begin() );
        std::copy( spec_in.begin()+1, spec_in.end()-1, spec_reflected.rbegin() );
        
        // transform to halfcomplex interferogram
        fft(spec_reflected, short_ifg);
        
        // copy ends to larger halfcomplex interferogram vector.
        std::copy( short_ifg.begin(), short_ifg.begin() + SPECTRUM_SIZE,
              ifg_zerofilled.begin() );
        std::copy( short_ifg.rbegin(), short_ifg.rbegin() + SPECTRUM_SIZE - 2,
              ifg_zerofilled.rbegin() );
        ifg_zerofilled[SPECTRUM_SIZE-1] *= 0.5; // hack to overcome a possible rfftw (FFTW2) bug.
        // FIXME - check if above still holds for FFTW3

        // re-transform to upsampled spectrum
        fftback(ifg_zerofilled, spec_upsampled_reflected);
        
        std::transform( spec_upsampled_reflected.begin(),
                   spec_upsampled_reflected.begin() + UPSAMPLED_SPECTRUM_SIZE,
                   spec_upsampled.begin(),
                   bind2nd(std::multiplies<double>(), NORMALIZATION) );
    } // upsample
    
    
    
    /***********************************
    downsample high-resolution intermediate spectrum
    ***********************************/    
    void linear_downsample( std::vector<double>& spec_out, 
                                         const std::vector<double> &spec_upsampled,
                                         const double vlaser_ratio )
    {
        double A = 32.0 * ( SPECTRUM_SIZE - 1 ) * ( 1 / vlaser_ratio - 1 );
        double B = 32.0 / vlaser_ratio;
        double r, rf; // fractional resample position
        int ri; // integral resample position
        for (unsigned int n = 0; n < SPECTRUM_SIZE; ++n )
        {
            r = A + B * n;
            ri = (int)floor(r);
            rf = r - ri;
            
            if (r < 0 || r >= UPSAMPLED_SPECTRUM_SIZE)
                spec_out[n] = 0;
            else
                spec_out[n] = spec_upsampled[ri]
                    + (spec_upsampled[ri+1] - spec_upsampled[ri]) * rf;
        }
    } // linear_downsample
    
    
public:
    void operator()(const std::vector<double>& obsIn, std::vector<double>& obsOut, double vlaser_ratio)
    {
	upsample( spec_upsampled, obsIn);
	
	// FIXME the vlaser_ratio should be computed from (computed or measured) theta in the GIFTS case
	
	linear_downsample( obsOut, spec_upsampled, vlaser_ratio );
    }
};

} // namespace

#endif
