/** 
*  \file SimCal1Test.cc
*  \brief Unit test for calibration for simulation data
*    
*  \author R.K.Garcia <rayg@ssec.wisc.edu>
*
*  \version $Id: SimCal1Test.cc,v 1.15 2005/11/22 03:17:43 maciek Exp $
*
*  \par Copyright:
*  \verbatim
*
*  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
*  Space Science and Engineering Center
*  University of Wisconsin - Madison, USA
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* FUTURE: consider parallelizing file I/O. Possibly group readers and 
* writers separately within a Binder Group and specify no sharing of 
* resources among different binders to prevent race conditions. Binders
* can themselves be multi-threaded if desired.
*
*  \endverbatim
*/



#include <cassert>
#include <list>
#include <algorithm>

#include "Types.h"
#include "Physics.h"  //wnums_t
#include "Planck.hxx"

#include "Cifg2Spectrum_GIFTSv1.hxx" // convert interferograms to spectra
    // FIXME: gotcha! the Cifg2Spectrum.hxx contains a virtual base class that is not abstract!
    // don't even need a virtual base class unless we really need a runtime factory..
#include "PiecewiseLinearCalibrationContext.hxx" 
#include "RadiometricCalibration.hxx"
#include "BindCdf.hxx"  // bring in source data and reference settings

using namespace gips;

class GIFTS_LW_Simulation_Traits_200511
{
public:
    // types which are used to represent the major buffer types throughout the unfolded system.
    typedef double RawIfgType; // as it arrives from the outside world
    typedef std::complex<double> IfgType; // as preferred to be manipulated
    typedef std::complex<double> SpcType; // uncalibrated spectrum
    typedef double CalSpcType; // calibrated spectrum, non-complex
    typedef double CoeffType;
    
    // compile-time constant sizes of these buffers
    enum 
    { 
        RawIfgSize = 2048, // aka rawInterferogramSize
        IfgSize = 2048, // aka shiftedInterferogramSize
        FftBufferSize = 2048, // aka doubledRawSpectrumSize
        
        // filtered and decimated spectrum sizes
        // FIXME: separate into SMW and LW
        SpcSize = 1024, // aka rawSpectrumSize
        CalSpcSize = 1024
    };

    static physics::wnums_t wavenumbers() // from ROK, wavenumbers for spectral form; move to CDF reference file
    {
        const double vlaser = 9398.0; 
        const double r_lw = 8; // reduction factor
        const double n_lw = IfgSize;
        
        double x_lw = (n_lw/2.0) * r_lw/vlaser;
        double dv = 1.0 / (2.0 * x_lw);
        double v0_lw = (n_lw/2 + 1) * dv;
        physics::wnums_t ret; // FIXME: make a friendlier constructor
        ret.baseWnum = v0_lw;
        ret.deltaWnum = dv;
        ret.wnumCount = SpcSize;
        return ret;
    }
};


typedef GIFTS_LW_Simulation_Traits_200511 myTraits;

typedef BufferTypes<myTraits> Types;



// quick and dirty controlling state machine
// FIXME: better separation of roles so that we can parallelize it and have multiple workers/pipelines
//
// interface description: 
// obs_in->scandir
// obs_in->scenetype
// obs_in->detector
//
// subsys->inputsvc?
//
// subsys->ifg2spc
// subsys->calstudy : invocable calstudy capsule
// subsys->calsynth
// subsys->calapply 
//
// subsys->outputsvc?

template< typename Traits, typename Ports >
struct TestBatchController: private Ports
{
    TestBatchController(Ports p_): Ports(p_)
    {
    }

    void studyPhase()
    {
        subsys->inputsvc.rewind();
        
        // read through blackbodies and load context cache
        do
        {
            switch( obs_in->scenetype )
            {
                case 'H':
                case 'A':
                case 'Z':
                    subsys->ifg2spc();
                    subsys->calstudy();
                    break;
                case 'E':
                    break;
                default:
                    std::cerr << "ERROR: scene type is " << obs_in->scenetype << endl;
            }
        } while( 0 == subsys->inputsvc.next() );        
    }

    void applyPhase()
    {
        subsys->inputsvc.rewind();
        
        do 
        {
            switch( obs_in->scenetype )
            {
                case 'H':
                case 'A':
                case 'Z':
                    break;
                case 'E':
                    subsys->ifg2spc();
                    subsys->calsynth();
                    subsys->calapply();
                default:
                    std::cerr << "ERROR: scene type is " << obs_in->scenetype << endl;
            }
        } while( 0 == subsys->inputsvc.next() );        
    }

    
    void operator()( void )
    {
        studyPhase();
        applyPhase();
    }
};


// input service feeds sufficient metadata (scene type et al) to system controller to allow it to 
// determine course of action; also feeds interferogram data parcels to worker/s.
// In this case, we feed through all the blackbodies first, followed by all the earth observations.
// usage: ctor push_back+ ( rewind next+ )+ dtor
class TestInputService
{
    std::list< boost::shared_ptr< NcFile > > files;
    std::list< FileToMemBinderGroup > all;
    std::list< FileToMemBinderGroup >::iterator cursor;
    
public:

    // add a binding set to the list; 
    void push_back( boost::shared_ptr<NcFile> file, const FileToMemBinderGroup &filebindings )
    {
        files.push_back( file );
        all.push_back( filebindings );
        cursor = all.begin();
    }

    void rewind()
    {
        cursor = all.begin();
        // read the data into memory
        cursor->update(0,SEEK_SET); 
    }
    
    // go on to the next record in the current file; if none remain, go to next file
    // returns 0 on success, <0 on failure (e.g. end-of-files)
    // assumes every file has at least one record
    int next()
    {
        try 
        {
            cursor->update(+1,SEEK_CUR); // FUTURE: consider using return code rather than exceptions; 0==success
            return 0;
        } 
        catch( FileToMemBinder::seekError se ) // one of the binders hit EOF
        { // change files
            cursor++;
            if (cursor == all.end()) return -1;
            cursor->update(0,SEEK_SET);
            return 0;
        }
        catch( FileToMemBinder::ioError ioe )
        {
            std::cerr << "Warning: ioError from binder group\n"; // FIXME: don't know if NetCDF EOF causes seekError or ioError
            cursor++;
            if (cursor == all.end()) return -1;
            cursor->update(0,SEEK_SET);            
            return 0;
        }
    }
};



// Data is loaded into the test bench, which is connected to the assortment of components needed
// to achieve the processing goal. Each component has limited visibility of the work area, with
// sub-structures satisfying their connection requirements. This reduces copying and dynamic allocation
// substantially, and permits "syntax-compatible" data containers to be used according to the situation;
// for instance, concurrency-controlled containers could be considered in the case that the
// components (e.g. the input service) are operating in multiple threads.

struct TestBench
{    
    typedef physics::wnums_t wavenumbers_t; // strongly consider moving this to traits
    typedef std::vector<double> emissivity_t;    // as well as this
    
    typedef PolynomialCalibrationContext<myTraits> ContextType;
    typedef ContextType::BasicCacheType CacheType;

    struct raw_observation
    {
        Types::RawInterferogram interferogram;
    }
        raw;
    
    struct uncal_observation
    {
        Types::Spectrum cplxspectrum; 
        double time;
        char scandir;
        char scenetype;
        int detector;
    } 
        uncal;

    struct calibration_context // is updated by the calstudy element and used by the calsynth
    {
        CacheType context_cache;
    }
        contexts;

    struct synthetic_blackbody  // is created by the calibration model using the context cache
    {
        Types::Spectrum cplxspectrum;
        double bbtemp;
        double reflectedtemp;
    } 
        zbb,
        abb,
        hbb;

    struct output_data
    {
        Types::CalSpectrum spectrum;
    } 
        cal,
        diag;      
    
    // reference data for longwave GIFTS simulation data
    // later, move this into a CDF file and bind it
    struct reference_data
    {    
        /////
        // interferogram conversion reference info
        unsigned int chopRawSpectrumFrom; 
        unsigned int chopRawSpectrumTo;
        bool wrapRawSpectrumLastPoint;

        /////
        // radiometric calibration info
        wavenumbers_t wavenumbers;
        emissivity_t bbemissivity; // set to 1.0 for the time being; later, fill with interpolated emissivity.txt@wavenumbers
        double telescopeTau, calmirrorTau;         // for the moment, 1.0 transmission
           
        reference_data(): // FIXME: get this all from CDF file using BindCdf
            chopRawSpectrumFrom(1024),
            chopRawSpectrumTo(2048),
            wrapRawSpectrumLastPoint(0),
            wavenumbers( myTraits::wavenumbers() ), 
            bbemissivity( wavenumbers.size(), 1.0 ), 
            telescopeTau(1.0), calmirrorTau(1.0) 
        {
        }
    } 
        ref;

    
    // FIXME: unify the alphabet_soup ofStudlyCaps VersusUberStudlyCaps at least by explaining convention used
    



    
    /////////////////////////////////////////////
    // Create a binder group for loading simulation data into input areas
    
    /// record mapping moves us pixel-by-pixel through the input files
    // FIXME: this record mapping needs to update an I,J pixel index as an ancillary input 
    // in the uncal_observation structure! 
    // The pixel number will be used as part of the key for the calibration context cache.
    GiftsPixelRecordMapping pixels_as_records;
    
    // the test input service iterates across a series of CDF input files, in sequence, 
    // pulling test data from each pixel in turn, and placing it in the workspace buffers.
    // A more advanced input service would transfer the data to the worker nodes, and 
    // require a more intelligent piece of system control logic, of course..
    TestInputService inputsvc;
    
    // FIXME: this needs to get some kind of time field from the input file! 
    void bindEarthFile( const std::string &filename )
    {
        boost::shared_ptr<NcFile> myfile( new NcFile( filename.c_str() ) ); 
        
        FileToMemBinderGroup grp;
        long shape[] = { 1, 1, Types::RawIfgSize };
        grp.push_back( newCdfVectorReader( &raw.interferogram, myfile->get_var("signalReal"), 
                        shape, shape+3, &pixels_as_records ) );
        inputsvc.push_back( myfile, grp );
    }
    
    // blackbody files are used during the study phase when we're in batch mode..
    // earth files are used in the apply phase. 
    // for a full GIFTS pipeline, the study and apply are interleaved (similar to S-HIS) with a 
    // predictive calibration modeler providing context to the system.
    void bindBlackbodyFile( const std::string &filename )
    {
        bindEarthFile(filename);
    }
    
    // binding the reference file includes information that was ground-characterized 
    // such as transmissivity, emissivity and responsivity coefficients.
    void bindReferenceFile( const std::string &filename )
    {
        //FileToMemBinderGroup grp;
    }
    
    
    
    
    
    ///////////////////////////////////////////// 
    // Okay, we've defined a set of buffers which we now want algorithms capsules to shuffle data between.
    // Let's use simple pointers as our connecitng agency, and hook the ports on the capsule instances
    // to pieces of data that are relevant to their operation.
    

    ////////////////////////////////////////////////////////////

    // this declares a structure we can use to connect in an ifg conversion 
    struct Ifg2SpcPorts  // FUTURE: consider templating this, parameterized on delivered struct types
    {
        typedef const reference_data *settings_p;
        typedef const raw_observation *obs_in_p;
        typedef uncal_observation *obs_out_p;
        
        settings_p settings;
        obs_in_p obs_in;
        obs_out_p obs_out;
        // FIXME: don't like the fact that it's easiest to just use the same struct on input as output
        // with different fields accepting the output. Alternative is copying or aliasing. Other alternatives?
    };

    // return a ports structure which can be used to initialize an ifg to spectrum conversion
    Ifg2SpcPorts ifg2spcPorts()
    {
        Ifg2SpcPorts ret = { &ref, &raw, &uncal };
        return ret;
    }

    ////////////////////////////////////////////////////////////

    // this declares a structure connecting a piecewise linear calibration study 
    struct CalStudyPorts
    {
        const uncal_observation *obs_in;
        calibration_context *cache_out;
    };
    
    CalStudyPorts calstudyPorts()
    {
        CalStudyPorts ret = { &uncal, &contexts };
        return ret;
    }
    
    ////////////////////////////////////////////////////////////

    // connection to a piecewise linear calibration synthesizer
    struct CalSynthPorts
    {
        const calibration_context *cache_in;
        const uncal_observation *obs_in;
        synthetic_blackbody *abb_out, *hbb_out, *zbb_out;
    };

    CalSynthPorts calsynthPorts()
    {
        CalSynthPorts ret = { &contexts, &uncal, &abb, &hbb, &zbb };
        return ret;
    }

    ////////////////////////////////////////////////////////////

    // this declares a structure we can use to connect in a radiometric calibration capsule
    struct RadCalPorts
    {
        typedef TestBench::wavenumbers_t wavenumbers_t; // violation of DRY?? alternatives?
        typedef TestBench::emissivity_t emissivity_t;    

        const synthetic_blackbody *abb_in,*hbb_in,*zbb_in;
        output_data *cal_out, *diag_out;
        const uncal_observation *obs_in;
        const reference_data *ref_in;
    };
    
    RadCalPorts radcalPorts()
    {
        RadCalPorts ret = { &abb, &hbb, &zbb, &cal, &diag, &uncal, &ref };
        return ret;
    }

    ////////////////////////////////////////////////////////////

    // Ifg2Spc converts the interferograms to properly formed spectra matching reference wavenumber scale
    typedef Ci2sImpl<myTraits,Ifg2SpcPorts> Ifg2SpcCapsule;     // FIXME: better (more meaningful) name
    
    // CalStudy takes blackbody observations and builds fit models (contexts) to interpolate/extrapolate from
    // until the predictive calibration study capsule is ready, we run in batch mode and separate the study 
    // and apply phases.
    typedef PiecewiseLinearCalibrationStudy<myTraits,CalStudyPorts> CalStudyCapsule;
    
    // CalSynth does the actual interpolation to create synthetic reference spectra
    typedef ThreeBodyPolynomialCalibrationSynthesis<myTraits,CalSynthPorts> CalSynthCapsule;
    
    // CalApply takes the earth observations and synthetic blackbody spectra to make absolute earth radiances
    typedef ThreeBodyRadiometricCalibration<myTraits,RadCalPorts> CalApplyCapsule;    
    
    // Control state machine cycles the sub-capsules according to observation information,
    // including driving I/O on the input and output services
    // typedef ....
    
    // ADD: I/O services, in the short term just throw together a CDF binder group to feed the test bench
    Ifg2SpcCapsule ifg2spc;
    CalStudyCapsule calstudy;
    CalSynthCapsule calsynth;
    CalApplyCapsule calibrate;
    // ADD: controlling state machine ... 
    
    
    // the test bench itself is ultimately mature into a full pipeline capsule. 
    TestBench():
        ifg2spc( ifg2spcPorts() ),
        calstudy( calstudyPorts() ),
        calsynth( calsynthPorts() ),
        calibrate( radcalPorts() )
    {
    }

};



// a quick test routine to fill a spectrum buffer with a planck curve
void setToPlanck( Types::Spectrum &S, double kelvin, const physics::wnums_t &wn )
{
    PlanckRadiances<physics::wnums_t> B(kelvin,&wn);
    std::copy( B.begin(), B.end(), S.begin() );
}

void setToPlanck( TestBench::synthetic_blackbody &S, double kelvin, const physics::wnums_t &wn )
{
    setToPlanck(S.cplxspectrum,kelvin,wn);
    S.bbtemp = kelvin;
}

// initialize a test bench with a numerically reasonable but extremely simple test case
void setupSimple( TestBench &tb )
{
    // create synthetic spectra to be calibrated against
    setToPlanck( tb.abb, 273.15, tb.ref.wavenumbers );
    setToPlanck( tb.hbb, 333.0, tb.ref.wavenumbers );
    setToPlanck( tb.zbb, 2.7 /*FIXME*/, tb.ref.wavenumbers );

    setToPlanck( tb.uncal.cplxspectrum, 300.0, tb.ref.wavenumbers );
    tb.uncal.time = 1.0;
    
}


// test1: exercise the calibrate unit
void test1()
{
    TestBench mybench;
    
    setupSimple( mybench );

    FileToMemBinderGroup shifter;
    SimpleCdfRecordMapping mapping;    
    NcFile fp( "testout.nc", NcFile::Write ); // as defined by testout.cdl
        
    long shape[] = { 1, myTraits::SpcSize };
    shifter.push_back( newCdfVectorWriter( &mybench.cal.spectrum, NcVar_p(fp.get_var("signalReal")), shape, shape+2, &mapping ));
    shifter.push_back( newCdfVectorWriter( &mybench.diag.spectrum, NcVar_p(fp.get_var("signalImag")), shape, shape+2, &mapping ));

    // move the binding to first record
    shifter.update(0,SEEK_SET);

    // quick radiometric calibration test
    mybench.calibrate();
    
    // shift to next record, writing record 0
    shifter.update(1,SEEK_SET);
}

template< typename Cv, typename Rv >
void demuxComplex( Rv &real, Rv& imag, const Cv &cplx )
{
    assert( cplx.size() == Types::SpcSize );
    assert( real.size() == Types::SpcSize );
    assert( imag.size() == Types::SpcSize );
    for( unsigned i=0; i<cplx.size(); i++ )
    {
        real[i] = cplx[i].real();
        imag[i] = cplx[i].imag();
    }   
}

void test_bind_intermediate_outputs( NcFile &fp, TestBench &mybench, FileToMemBinderGroup &shifter, SimpleCdfRecordMapping &mapping )
{
    long rawshape[] = { 1, myTraits::RawIfgSize };
    shifter.push_back( newCdfVectorWriter( &mybench.raw.interferogram, NcVar_p(fp.get_var("raw")), 
                        rawshape, rawshape+2, &mapping ));
        
    // map the output file, linear records containing spectral real and imag components
    long shape[] = { 1, myTraits::SpcSize };
    shifter.push_back( newCdfVectorWriter( &mybench.cal.spectrum, NcVar_p(fp.get_var("uncalReal")), 
                        shape, shape+2, &mapping ));
    shifter.push_back( newCdfVectorWriter( &mybench.diag.spectrum, NcVar_p(fp.get_var("uncalImag")), 
                        shape, shape+2, &mapping ));
}








// test 1a: exercise ifgshift

struct cplxifgout 
{
    Types::Interferogram cplxinterferogram; // output of ifgshift
};

struct tmpports
{
    const TestBench::raw_observation *obs_in;
    cplxifgout *obs_out;
} ports;

struct getreal // unary functor
{ inline double operator()( const std::complex<double> &x ) const { return x.real(); } };

void test1a()
{
    TestBench mybench;
    cplxifgout result;
    
    ports.obs_in = &mybench.raw;
    ports.obs_out = &result;

    // set up test output to CDF    
    SimpleCdfRecordMapping mapping;    
    FileToMemBinderGroup shifter;
    NcFile fp( "testout.nc", NcFile::Write ); // as defined by testout.cdl
    test_bind_intermediate_outputs( fp, mybench, shifter, mapping );

    // map the input file, each effective record being a different pixel
    mybench.bindEarthFile( std::string("testin.nc") );
    
    // pull in the first record from the input
    mybench.inputsvc.rewind();
    // move the output binding to first record
    shifter.update(0,SEEK_SET);

    // build capsule and run it
    RealToCplxIfgShift<myTraits,tmpports> ifgshift(ports);
    ifgshift();
    
    // copy the results into raw so that it gets caught by the disk
    std::transform( result.cplxinterferogram.begin(), 
                    result.cplxinterferogram.end(), 
                    mybench.raw.interferogram.begin(), getreal() );
    
    shifter.update(+1,SEEK_CUR);
};

struct tmpports_fft
{
    const cplxifgout *obs_in;
    TestBench::uncal_observation *obs_out;
};

// test ifgshift and cplxfft
void test1b()
{
    TestBench mybench;
    cplxifgout result;
    
    ports.obs_in = &mybench.raw;
    ports.obs_out = &result;

    // set up test output to CDF    
    SimpleCdfRecordMapping mapping;    
    FileToMemBinderGroup shifter;
    NcFile fp( "testout.nc", NcFile::Write ); // as defined by testout.cdl
    test_bind_intermediate_outputs( fp, mybench, shifter, mapping );

    // map the input file, each effective record being a different pixel
    mybench.bindEarthFile( std::string("testin.nc") );
    
    // pull in the first record from the input
    mybench.inputsvc.rewind();
    // move the output binding to first record
    shifter.update(0,SEEK_SET);

    // build capsule and run it
    RealToCplxIfgShift<myTraits,tmpports> ifgshift(ports);
    ifgshift();
    
    tmpports_fft ports_fft = { &result, &mybench.uncal };
    
    // fft the complex ifg back into the workspace where it'll be picked up as uncal
    CplxFFT<myTraits,tmpports_fft> fft(ports_fft);
    fft();
    
    // copy the real part of the ifgshift results into raw so that it gets caught by the disk
    std::transform( result.cplxinterferogram.begin(), 
                    result.cplxinterferogram.end(), 
                    mybench.raw.interferogram.begin(), getreal() );

    // manually rip it apart for CDF prep        
    demuxComplex( mybench.cal.spectrum, mybench.diag.spectrum, mybench.uncal.cplxspectrum );
    
    shifter.update(+1,SEEK_CUR);
};






// test2: exercise ifg2spc
void test2()
{
    TestBench mybench;
    
    setupSimple( mybench );

    FileToMemBinderGroup shifter;

    // map the input file, each effective record being a different pixel
    mybench.bindEarthFile( std::string("testin.nc") );
            
    SimpleCdfRecordMapping mapping;    
    NcFile fp( "testout.nc", NcFile::Write ); // as defined by testout.cdl
    test_bind_intermediate_outputs( fp, mybench, shifter, mapping );

    // pull in the first record from the input
    mybench.inputsvc.rewind();
    // move the output binding to first record
    shifter.update(0,SEEK_SET);

    for(int i=0;i<5;i++) 
    {
        // quick ifg2spc test
        mybench.ifg2spc();
        
        // manually rip it apart for CDF prep
        demuxComplex( mybench.cal.spectrum, mybench.diag.spectrum, mybench.uncal.cplxspectrum );
        
        std::cout << mybench.cal.spectrum[0] << std::endl;
        std::cout << mybench.diag.spectrum[0] << std::endl;
        
        // shift to next record, writing record 0
        shifter.update(+1,SEEK_CUR);
        // advance input data to next available record
        mybench.inputsvc.next();
    }
}

// test3: exercise calstudy
void test3( int argc, char *argv[] )
{
    TestBench mybench;
    
    // setupSimple( mybench );

    // bind a series of blackbody interferogram files to the bench input service
    // and run ifg2spc and calstudy on each of the records. 
    
    for( int i=0; i<argc; i++ )
        mybench.bindEarthFile( std::string(argv[i]) );
    
    mybench.inputsvc.rewind();

    std::cout << "read-through check..\n";
    for( int i=0; 0==mybench.inputsvc.next(); i++ ) 
    {
        std::cout << "record " << i << "    \r";
        std::cout.flush();
    }
    
    mybench.inputsvc.rewind();
    std::cout << "\n\ncalstudy run-through..\n";
    
    double time_FIXME = 1.0;
    do
    {
        // FIXME: these should be provided by the input files!
        mybench.uncal.time = time_FIXME;
        time_FIXME += 1.0;
        mybench.uncal.scandir = 'F'; // no scan direction for the simulation data
        mybench.uncal.scenetype = 'H'; // FIXME: these are distinguished by the filename

        mybench.ifg2spc();
        mybench.calstudy();    
        
        std::cout << "elements in cache: " << mybench.contexts.context_cache.size() << std::endl;
    } 
    while( mybench.inputsvc.next() == 0 );
}

// test4: exercise calsynth
void test4()
{
}


int main( int argc, char *argv[] )
{
    test3(argc-1,argv+1);
}
