/**
 * \file SpectrumChop.h
 * \author Maciek Smuga-Otto <maciek@ssec.wisc.edu>
 *
 *  \version $Id: SpectrumChop.hxx,v 1.4 2005/10/28 01:09:02 rayg Exp $
 *
 *  \par Copyright:
 *  \verbatim
 *
 *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
 *  Space Science and Engineering Center
 *  University of Wisconsin - Madison, USA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  \endverbatim
 */

#ifndef H_SPECCHOP
#define H_SPECCHOP

#ifndef NO_DEPENDENCY_INCLUDES

#include <complex>
#include "Physics.h"
#include "Instrument.h"
#include "ProtocolRole.h"
#include "Operator.h"
#include "Types.h"

#endif

// NOTE: This file is an example of the kind of output expected from the 
// GIPS flanging compiler, and will in the future NOT be generated by hand.
// As a result, all the input/output structs are Plain Old Data (POD)s,
// as indicated by the _t suffixes. Furthermore, all their members have
// names such as would be published in the global data dictionary.


/* Interface specification

SpectrumChop<Traits,Ports>
    Input port: settings
        settings->wrapRawSpectrumLastPoint boolean, whether to wrap or not
        settings->chopRawSpectrumFrom starting offset (zero-based)
        settings->chopRawSpectrumTo endpoint (noninclusive) 
    Input port: obs_in
        obs_in->cplxspectrum, std::vector<> semantics      
    Output port: obs_out
        obs_out->cplxspectrum, std::vector<> semantics
*/

namespace gips 
{


template< typename Traits, typename Ports >
class SpectrumChop 
{
protected:
    typedef BufferTypes<Traits> Types;
    Ports P;

public:    
    explicit SpectrumChop(const Ports &p): P(p)
    {    
        // check consistency of symbols defined in settings. Namely:
        //assert( Types::SpectrumSize == settings.chopRawSpectrumTo - settings.chopRawSpectrumFrom + 1 );
    }
    
    Operator::Result operator()(void)
    {    
        // yes, I'm being a bad boy and copying data. boo hoo hoo.
        // convert to std::copy_n with iterators?
        for (unsigned int src(P.settings->chopRawSpectrumFrom), dest(0); 
                src < P.settings->chopRawSpectrumTo; 
                ++src, ++dest)
        {
            P.obs_out->cplxspectrum[dest] = P.obs_in->cplxspectrum[src];
        };
        
        // curious case of the last point: wrap it, or zero?
        P.obs_out->cplxspectrum[P.obs_out->cplxspectrum.size() - 1] = 
            P.settings->wrapRawSpectrumLastPoint ? P.obs_in->cplxspectrum[P.settings->chopRawSpectrumFrom] :
                                                   0.0;
    
        //cerr << "SpectrumChop: dataOut@600: " << obs_out.rawSpectrum[600] << endl; // DEBUG 
    
        return Operator::OK;
    }

    ~SpectrumChop() { }
}; // class SpectrumChop
        
        
template< typename Traits >
class SpectrumChopInterface
{
public:
    typedef BufferTypes<Traits> Types;

    struct settings_t 
    {
        unsigned int chopRawSpectrumFrom; 
        unsigned int chopRawSpectrumTo;
        char wrapRawSpectrumLastPoint;
    };
    
    struct obs_in_t 
    {
        typename Types::FftSpectrum cplxspectrum;
    };
    
    struct obs_out_t 
    {
        typename Types::Spectrum cplxspectrum;
    };
    
    struct Ports {
        typedef const settings_t *settings_p;
        typedef const obs_in_t *obs_in_p;
        typedef obs_out_t *obs_out_p;
        
        settings_p settings;
        obs_in_p obs_in;
        obs_out_p obs_out;
        
        explicit Ports(         
            settings_p settings_,
            obs_in_p obs_in_,
            obs_out_p obs_out_
        ):
               settings(settings_),
               obs_in(obs_in_),
               obs_out(obs_out_)
        {
        }
    };
};

        
} // namespace


#endif
