/**
 * \file Types.h
 * \brief template expansions used to declare in-memory representations and protocols
 *
 * \author Raymond K. Garcia <rayg@ssec.wisc.edu>
 *
 *
 *  \version $Id: Types.h,v 1.4 2005/11/05 22:38:24 rayg Exp $
 *
 *  \par Copyright:
 *  \verbatim
 *
 *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
 *  Space Science and Engineering Center
 *  University of Wisconsin - Madison, USA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *  \endverbatim
 *
 */

#ifndef H_GIPS_TYPES
#define H_GIPS_TYPES

#ifndef NO_DEPENDENCY_INCLUDES

#include <Physics.h>

#include <complex>

#include <boost/array.hpp>

#endif


// Traits defines the following:
// RawInterferogram, RawInterferogramSize, RawInterferogramType
// Interferogram, InterferogramSize, InterferogramType
// Spectrum, SpectrumSize, SpectrumType
// 
// We're going to focus on making the algorithm capsules into templates, 
// hinged on traits classes which compose at compile time to provide 
//  -- base buffer types and sizes (possibly max buffer sizes if we have run-time variation) 
//  -- protocols 
// 
// We use boost-style template composition in part to make name-based binding instead of 
// positional binding.
//
// This sets us up for automatic generation of traits classes including connector protocols
// without munging the human-written source.
/* an example of an unfolded traits class which would be used to bind a capsule to its protocols
class Ifg2Spectrum_instance1
{
    public:
    // protocol declaration for PortA on the 
    struct PortA
    {
        ... fields from data dictionary with appropriate types ...
    };
    
    struct PortB
    {
        ... 
    };
}
*/
/* In practice and for test code, mix-and-match of component protocols happens using template composition (aggregation?) syntax like
template< typename > TestProto1;
template< typename > TestProto2;
struct null_class { }; // see if boost includes this in its standard
class TestProtoSet: public TestProto1< TestProto2< null_class > > { };

int main()
{ 
    // bind the capsule to its protocols, buffer types and sizes
    MyCapsule< TestProtoSet > smw_ifg;
}
*/




// TO BE PROMOTED (to a GIFTS.h file?) - placed here for expediency
class GIFTS_Unfiltered_Traits_20050930
{  
public:
    // types which are used to represent the major buffer types throughout the unfolded system.
    typedef int16_t RawIfgType;
    typedef int16_t IfgType;
    typedef std::complex<double> SpcType;
    typedef double CalSpcType;
    typedef double CoeffType;
    
    // compile-time constant sizes of these buffers
    enum 
    { 
        RawIfgSize = 32768,
        RawSpcSize = 32768,
        FftBufferSize = 32768,
        
        // filtered and decimated spectrum sizes
        LwSpcSize = 2049,
        SmwSpcSize = 4097
    };
};    

class GIFTS_Test_Traits_200508
{  
public:
    // types which are used to represent the major buffer types throughout the unfolded system.
    typedef std::complex<double> RawIfgType; // as it arrives from the outside world
    typedef std::complex<double> IfgType; // as preferred to be manipulated
    typedef std::complex<double> SpcType; // uncalibrated spectrum
    typedef double CalSpcType; // calibrated spectrum, non-complex
    typedef double CoeffType;
    
    // compile-time constant sizes of these buffers
    enum 
    { 
        RawIfgSize = 2048, // aka rawInterferogramSize
        IfgSize = 2048, // aka shiftedInterferogramSize
        FftBufferSize = 2048, // aka doubledRawSpectrumSize
        
        // filtered and decimated spectrum sizes
        // FIXME: separate into SMW and LW
        SpcSize = 2048, // aka rawSpectrumSize
        CalSpcSize = 2048
        //SmwSpcSize = 1025
    };
};    

typedef GIFTS_Test_Traits_200508 Test_Traits;

// given instrument dimensionality and base representation traits, start turning it into buffer types
template < typename Base >
class BufferTypes: public Base
{
    public:
    template< std::size_t n > class RawInterferogramBuffer: 
        public boost::array< typename Base::RawIfgType, n > 
        { };
    template< std::size_t n > class InterferogramBuffer: 
        public boost::array< typename Base::IfgType, n > 
        { };
    template< std::size_t n > class SpectrumBuffer: 
        public boost::array< typename Base::SpcType, n > 
        { };
    template< std::size_t n > class CalSpectrumBuffer: 
        public boost::array< typename Base::CalSpcType, n > 
        { };
    template< std::size_t n > class CoeffBuffer: 
        public boost::array< typename Base::CoeffType, n > 
        { };
    typedef RawInterferogramBuffer< Base::RawIfgSize> RawInterferogram;
    typedef InterferogramBuffer<Base::IfgSize> Interferogram;
    typedef SpectrumBuffer<Base::FftBufferSize> FftSpectrum;
    typedef CoeffBuffer<Base::SpcSize> CoeffSpectrum;
    typedef SpectrumBuffer<Base::SpcSize> Spectrum;
    typedef CalSpectrumBuffer<Base::SpcSize> CalSpectrum;
    enum 
    { 
        SpectrumSize = Base::SpcSize, 
        InterferogramSize = Base::IfgSize, 
        RawInterferogramSize = Base::RawIfgSize,
        FftBufferSize = Base::FftBufferSize,
        CalSpectrumSize = Base::CalSpcSize
    };
};


#endif

